#
# Copyright (c) 2022 LAAS/CNRS
# All rights reserved.
#
# Redistribution  and  use  in  source  and binary  forms,  with  or  without
# modification, are permitted provided that the following conditions are met:
#
#   1. Redistributions of  source  code must retain the  above copyright
#      notice and this list of conditions.
#   2. Redistributions in binary form must reproduce the above copyright
#      notice and  this list of  conditions in the  documentation and/or
#      other materials provided with the distribution.
#
# THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
# WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
# MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
# ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
# WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
# IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#
#                                           Anthony Mallet on Wed Jan  5 2022
#
import time
import warnings

from . import event


# === Handle ===============================================================
#
class Handle:
  """Status and result of a request."""

  # --- __init__ -----------------------------------------------------------
  #
  def __init__(self, server, token, rid):
    """Initialize self. Instances are created as the result of a request
    invocation by a component client. Manually created instances will
    serve no purpose.
    """
    self.__server = server
    self.__token = token
    self.__rid = rid
    self.__info = {}
    self.__callback = None


  # --- callback -----------------------------------------------------------
  #
  @property
  def callback(self):
    """The callback to be run for each status update.

    When setting this, the new callback may be invoked once for all past
    events if no callback was installed previously. The callback will be
    invoked once for each future event.
    """
    return self.__callback

  @callback.setter
  def callback(self, cb):
    if cb is not None:
      if not callable(cb):
        raise TypeError('callback must be callable')

      # run callback if pending
      if self.__callback is ...: cb(self)

    self.__callback = cb


  # --- status -------------------------------------------------------------
  #
  @property
  def status(self):
    """Current status of the request as a genomix.Status object."""
    return event.Status(self.__info.get('status', None))


  # --- result -------------------------------------------------------------
  #
  @property
  def result(self):
    """Output of the request as a dictionary.

    If request has the status 'error', a GenoMError exception is raised,
    with the details of the error. If request is not yet in status 'done',
    a RuntimeError exception is raised.
    """
    if self.status < event.Status.done:
      raise RuntimeError('request ' + (
        'not sent' if self.status < event.Status.sent else 'in progress'
      ))
    elif self.status == event.Status.error:
      raise event.GenoMError(self.__info.get('errmsg', {}))

    return self.__info.get('output', {})


  # --- abort --------------------------------------------------------------
  #
  def abort(self):
    """Request for cancellation of the request.
    If the request was already done, this does nothing.
    """
    if self.__server:
      self.__server.abort(self.__token, self.__rid)


  # --- wait ---------------------------------------------------------------
  #
  def wait(self, status = event.Status.done, *, timeout = None):
    """Block and process events until the request is in the expected status.
    If a timeout is given, this blocks for at most this amount of seconds.
    With a timeout of 0, this will not block but still process any pending
    events.

    While events are processed, callbacks will be invoked, including for
    unrelated requests.
    """
    if status == event.Status.none: return
    if status not in (event.Status.done, event.Status.sent):
      raise ValueError('bad status')

    if timeout is not None:
      deadline = time.monotonic() + timeout

    while self.status < status:
      if timeout is not None:
        timeout = deadline - time.monotonic()
      if event.mux.wait(timeout) is None:
        raise TimeoutError


  # --- __await__ ----------------------------------------------------------
  #
  def __await__(self):
    """Suspend the execution until the request is done. Events will be
    processed and callbacks may be invoked, including for unrelated
    requests.

    The returned value is the result attribute, or a GenoMError
    exception.
    """
    while self.status < event.Status.done:
      yield from event.mux.__await__()

    return self.result


  # --- _update ------------------------------------------------------------
  #
  def _update(self, *, server = ..., reply = None):
    """Update internal state"""
    if server is not ...: self.__server = server

    if reply:
      self.__info.update(reply)

      # run callback on status change
      if self.__callback not in (None, ...):
        self.__callback(self)
      else:
        # mark callback as runnable for when it is set later
        self.__callback = ...

      # clean-up remotely when done. If the server has closed in the meantime,
      # one can safely ignore the error.
      if self.__server and self.status >= event.Status.done:
        try:
          self.__server.clean(self.__token, self.__rid)
        except RuntimeError: pass


  # --- __repr__ -----------------------------------------------------------
  #
  def __repr__(self):
    """Returns a string with the component name, an internal request id
    and the status."""
    return '<%s::%s %s>' % (self.__token, self.__rid, self.status)


# === Handles ==============================================================
#
class Handles:
  """Set of request handles indexed by their unique id."""

  # --- __init__ -----------------------------------------------------------
  #
  def __init__(self):
    self.__active = {}


  # --- __call__ -----------------------------------------------------------
  #
  def __call__(self, server, token, rid):
    """Get an existing request instance or create one."""
    crid = (token, rid)
    r = self.__active.get(crid, None)
    if r is None:
      r = self.__active[crid] = Handle(server, token, rid)

    if server is not None:
      r._update(server = server)

    return r


  # --- close --------------------------------------------------------------
  #
  def close(self, token):
    """Remove all pending requests for a component."""

    for t, rid in [(t, rid) for t, rid in self.__active
                   if t == token or token is ...]:
      self.clean(t, rid)


  # --- clean --------------------------------------------------------------
  #
  def clean(self, token, rid):
    """Remove terminated request handle."""

    r = self.__active.pop((token, rid), None)
    if not r:
      raise RuntimeError('bad request id')

    # prevent further interaction with the server
    r._update(server = None)

    # cancel pending status
    if r.status < event.Status.done:
      r._update(reply = {
        'status': str(event.Status.error),
        'errmsg': {'ex': '::genom::interrupted',
                   'detail': {'by': 'local client'}}
      })


  # --- reply --------------------------------------------------------------
  #
  def reply(self, data):
    """Dispatch incoming requests replies"""

    # get request id
    crid = data.pop('id', None)
    if crid is None:
      warnings.warn(RuntimeWarning('dropping malformed reply'))
      return

    # get the handle instance: because of an expected race condition, it is
    # possible that a reply is received before the handle is created in the
    # sending function. Passing None as the server prevents the handle to be
    # cleaned before it is created there.
    token, _, rid = crid.partition('::')
    r = self(None, token, rid)

    # dispatch
    r._update(reply = data)
